#ifndef STATIC_CONTAINER_VECTOR_H

#define STATIC_CONTAINER_VECTOR_H

#include <algorithm>
#include <boost/assert.hpp>
#include "static_container/rotate.h"
#include "static_container/STATIC_CONTAINER_MEMBERTYPEDEF.h"
#include "static_container/destruct.h"
#include "static_container/compare_methods.h"
#include "static_container/integer.h"
#include <algorithm>
#include <boost/call_traits.hpp>

namespace static_container {

///	p}N
#define STATIC_VECTOR_FOREACH( op )	\
	for ( iterator it = begin(); it != end(); ++it ) {\
		op;\
	}

	///	ÓÏ ( z )  std::vector ̂悤Ɉ߂̃NX
	/**
	 *	݁Af[^́A4 byte ACۏႳ܂B
	 *	SIMD pf[^Ȃ 4 byte ł̓ACȂꍇ́AĝŁA
	 *	ĂB
	 *	ƂẮAbuffer_ AC]ɂƂāA
	 *	s this āAKvItZbgݒ肵obt@𗘗p\łB
	 */
	template< typename Value, size_type MaxSize >
	class vector : public compare_methods< vector< Value, MaxSize > > {
	public:
		STATIC_CONTAINER_MEMBERTYPEDEF( Value )
		typedef typename boost::call_traits< Value >::param_type	param_type;
		typedef pointer			iterator;
		typedef const_pointer	const_iterator;
	private:
		size_type	size_;
		uint8_t		buffer_[ MaxSize * sizeof( Value ) ];
		
		static void destruct( reference v ) {
			static_container::destruct< Value >( v );
		}

		///	pB 擪vf̃AhX𓾂
		pointer	top() {
			return reinterpret_cast< pointer >( buffer_ );
		}

		///	pB 擪vf̃AhX𓾂
		const_pointer	top() const {
			return reinterpret_cast< const_pointer >( buffer_ );
		}
	public:
		///	őTCY
		BOOST_STATIC_CONSTANT( size_type, const_max = MaxSize );

		///	őTCY擾
		static size_type max_size() { return const_max; }

		///	eʎ擾
		static size_type capaciry() { return const_max; }

		/// TCY擾
		size_type size() const { return size_; }

		iterator begin() { return reinterpret_cast< iterator >( top() ); }
		const_iterator begin() const { return reinterpret_cast< const_iterator >( top() ); }
		iterator end() { return begin() + size(); }
		const_iterator end() const { return begin() + size(); }

		///	ɑ΂vfǉ
		void push_back( param_type v ) {
			BOOST_ASSERT( size() < max_size() );
			new( &top()[ size() ] ) Value( v );
			++size_;
		}

		///	vf폜
		void pop_back() {
			BOOST_ASSERT( false == empty() );
			--size_;
			destruct( top()[ size() ] );
		}

		/// ftHgRXgN^
		vector() : size_( 0 ) {
		}
		
		/// TCYwRXgN^
		/**
		 *	evfɑ΂ẮAftHgRXgN^Ăяo
		 */
		vector( size_type s ) : size_( s ) {
			BOOST_ASSERT( s <= max_size() );
			STATIC_VECTOR_FOREACH( new( it ) Value() )
		}
		/// TCYwATCRXgN^
		/**
		 *	evfɑ΂ẮARs[RXgN^Ăяo
		 */
		vector( size_type s, param_type v ) : size_( s ) {
			BOOST_ASSERT( s <= max_size() );
			STATIC_VECTOR_FOREACH( new( it ) Value( v ) )
		}
		/// Rs[RXgN^
		vector( const vector& other ) : size_( other.size_ ) {
			const_iterator otherIt = other.begin();
			STATIC_VECTOR_FOREACH( new( it ) Value( *( otherIt++ ) ) )
		}
		// fXgN^
		~vector() {
			clear();
		}
		/// Zq
		vector& operator = ( const vector& other ) {
			if ( this != &other ) {
				assign( other.begin(), other.end() );
			}
			return *this;
		}
		
		///	
		template < typename InputIt >
		void assign( InputIt first, InputIt last ) {
			clear();
			iterator dest = begin();
			for ( InputIt src = first; src != last; ++src ) {
				new( dest ) Value( *src );
				++size_;
			}
		}
		
		///	it 폜܂
		void erase( iterator it ) {
			rotate( it, it + 1, end() );
			pop_back();
		}

		///	[ first, last ) 폜܂
		/**
		 *	݁AɏdłB
		 *	v]΍܂B
		 */
		void erase( iterator first, iterator last ) {
			difference_type sizeDiff = std::distance( first, last );
			rotate( first, last, end() );
			for ( size_type i = 0; i < sizeDiff; ++i ) {
				pop_back();
			}
		}
		
		///	xNg
		void clear() {
			for ( iterator it = begin(); it != end(); ++it ) {
				destruct( *it );
			}
			size_ = 0;
		}

		Value& operator [] ( size_type i ) {
			PYD_ASSERT( i < size() );
			return *( begin() + i );
		}
		const Value& operator [] ( size_type i ) const {
			PYD_ASSERT( i < size() );
			return *( begin() + i );
		}
		bool empty() const {
			return 0 == size();
		}
		reference at( size_type i ) { return operator [] ( i ); }
		param_type at( size_type i ) const { return operator [] ( i ); }

		// front() and back()
		reference front() { return *begin(); }
		param_type front() const { return *begin(); }
		reference back() { return *( end() - 1 ); }
		param_type back() const { return *( end() - 1 ); }

		///	u
		void swap( vector& other ) {
			std::swap_ranges( begin(), end(), other.begin() );
		}
	    
		// assign one value to all elements
		void assign ( param_type value ) {
			std::fill_n( begin(), size(), value );
		}
	};
	#undef STATIC_VECTOR_FOREACH
}

#endif